@@ -3,32 +3,7 @@
 use anyhow::bail;
 use anyhow::Result;
 
-use crate::combinators::assert;
-use crate::combinators::assert_exists;
-use crate::combinators::char;
-use crate::combinators::check;
-use crate::combinators::check_not;
-use crate::combinators::delimited;
-use crate::combinators::if_true;
-use crate::combinators::many0;
-use crate::combinators::many_till;
-use crate::combinators::map;
-use crate::combinators::maybe;
-use crate::combinators::next_char;
-use crate::combinators::or;
-use crate::combinators::or3;
-use crate::combinators::or4;
-use crate::combinators::or5;
-use crate::combinators::preceded;
-use crate::combinators::separated_list;
-use crate::combinators::skip_whitespace;
-use crate::combinators::tag;
-use crate::combinators::take_while;
-use crate::combinators::terminated;
-use crate::combinators::with_error_context;
-use crate::combinators::ParseError;
-use crate::combinators::ParseErrorFailure;
-use crate::combinators::ParseResult;
+use crate::combinators::*;
 
 // Shell grammar rules this is loosely based on:
 // https://pubs.opengroup.org/onlinepubs/009604499/utilities/xcu_chap02.html#tag_02_10_02
@@ -50,12 +25,14 @@ pub enum Sequence {
   EnvVar(EnvVar),
   /// `MY_VAR=5`
   ShellVar(EnvVar),
-  // cmd_name <args...>
+  /// `cmd_name <args...>`
   Command(Command),
-  // cmd1 | cmd2
+  /// `cmd1 | cmd2`
   Pipeline(Box<Pipeline>),
-  // cmd1 && cmd2 || cmd3
+  /// `cmd1 && cmd2 || cmd3`
   BooleanList(Box<BooleanList>),
+  /// `(list)`
+  Subshell(Box<SequentialList>),
 }
 
 #[derive(Debug, Clone, PartialEq)]
@@ -156,9 +133,12 @@ impl StringOrWord {
 
 #[derive(Debug, PartialEq, Clone)]
 pub enum StringPart {
-  Variable(String),
+  /// Text in the string (ex. `some text`)
   Text(String),
-  SubShell(SequentialList),
+  /// Variable substitution (ex. `$MY_VAR`)
+  Variable(String),
+  /// Command substitution (ex. `$(command)`)
+  Command(SequentialList),
 }
 
 /// Note: Only used to detect redirects in order to give a better error.
@@ -210,10 +190,9 @@ fn parse_sequential_list(input: &str) -> ParseResult<SequentialList> {
     terminated(parse_sequential_list_item, skip_whitespace),
     terminated(
       skip_whitespace,
-      or3(
+      or(
         map(parse_sequential_list_op, |_| ()),
         map(parse_async_list_op, |_| ()),
-        check(char(')')),
       ),
     ),
   )(input)?;
@@ -232,9 +211,13 @@ fn parse_sequential_list_item(input: &str) -> ParseResult<SequentialListItem> {
 }
 
 fn parse_sequence(input: &str) -> ParseResult<Sequence> {
-  let (input, current) = or(
-    parse_env_or_shell_var_command,
-    map(parse_command, Sequence::Command),
+  let (input, current) = terminated(
+    or3(
+      map(parse_subshell, |l| Sequence::Subshell(Box::new(l))),
+      parse_env_or_shell_var_command,
+      map(parse_command, Sequence::Command),
+    ),
+    skip_whitespace,
   )(input)?;
   let (input, current) = match parse_boolean_list_op(input) {
     Ok((input, op)) => {
@@ -349,7 +332,7 @@ fn parse_boolean_list_op(input: &str) -> ParseResult<BooleanListOperator> {
 }
 
 fn parse_sequential_list_op(input: &str) -> ParseResult<&str> {
-  parse_op_str(";")(input)
+  terminated(tag(";"), skip_whitespace)(input)
 }
 
 fn parse_async_list_op(input: &str) -> ParseResult<&str> {
@@ -359,17 +342,18 @@ fn parse_async_list_op(input: &str) -> ParseResult<&str> {
 fn parse_op_str<'a>(
   operator: &str,
 ) -> impl Fn(&'a str) -> ParseResult<'a, &'a str> {
+  debug_assert!(operator == "&&" || operator == "||" || operator == "&");
   let operator = operator.to_string();
   terminated(
     tag(operator),
-    terminated(check_not(special_char), skip_whitespace),
+    terminated(check_not(one_of("|&")), skip_whitespace),
   )
 }
 
 fn parse_pipeline_op(input: &str) -> ParseResult<()> {
   terminated(
     map(char('|'), |_| ()),
-    terminated(check_not(special_char), skip_whitespace),
+    terminated(check_not(char('|')), skip_whitespace),
   )(input)
 }
 
@@ -400,7 +384,12 @@ fn parse_env_var(input: &str) -> ParseResult<EnvVar> {
 }
 
 fn parse_env_var_name(input: &str) -> ParseResult<&str> {
-  take_while(is_valid_env_var_char)(input)
+  let (input, name) = take_while(is_valid_env_var_char)(input)?;
+  if name.is_empty() {
+    ParseError::backtrace()
+  } else {
+    Ok((input, name))
+  }
 }
 
 fn parse_env_var_value(input: &str) -> ParseResult<StringOrWord> {
@@ -415,9 +404,10 @@ fn parse_string_or_word(input: &str) -> ParseResult<StringOrWord> {
 }
 
 fn parse_word(input: &str) -> ParseResult<Vec<StringPart>> {
+  const SPECIAL_CHARS: &str = "*~(){}<>$|&;\"'";
   assert(
     parse_string_parts(|c| {
-      !c.is_whitespace() && (c == '$' || !is_special_char(c))
+      !c.is_whitespace() && (c == '$' || !SPECIAL_CHARS.contains(c))
     }),
     |result| {
       result
@@ -468,7 +458,10 @@ fn parse_single_quoted_string(input: &str) -> ParseResult<&str> {
   delimited(
     char('\''),
     take_while(|c| c != '\''),
-    assert_exists(char('\''), "Expected closing single quote."),
+    with_failure_input(
+      input,
+      assert_exists(char('\''), "Expected closing single quote."),
+    ),
   )(input)
 }
 
@@ -478,7 +471,10 @@ fn parse_double_quoted_string(input: &str) -> ParseResult<Vec<StringPart>> {
   delimited(
     char('"'),
     parse_string_parts(|c| c != '"'),
-    assert_exists(char('"'), "Expected closing double quote."),
+    with_failure_input(
+      input,
+      assert_exists(char('"'), "Expected closing double quote."),
+    ),
   )(input)
 }
 
@@ -495,6 +491,25 @@ pub(crate) fn parse_string_parts(
     )(input)
   }
 
+  fn parse_special_shell_var(input: &str) -> ParseResult<char> {
+    // for now, these hard error
+    preceded(char('$'), |input| {
+      if let Some(char) = input.chars().next() {
+        // $$ - process id
+        // $? - last exit code
+        // $# - number of arguments in $*
+        // $* - list of arguments passed to the current process
+        if "$?#*".contains(char) {
+          return ParseError::fail(
+            input,
+            format!("${} is currently not supported.", char),
+          );
+        }
+      }
+      ParseError::backtrace()
+    })(input)
+  }
+
   fn parse_escaped_char<'a>(
     c: char,
   ) -> impl Fn(&'a str) -> ParseResult<'a, char> {
@@ -504,7 +519,8 @@ pub(crate) fn parse_string_parts(
   fn first_escaped_char<'a>(
     allow_char: impl Fn(char) -> bool,
   ) -> impl Fn(&'a str) -> ParseResult<'a, char> {
-    or4(
+    or5(
+      parse_special_shell_var,
       parse_escaped_dollar_sign,
       parse_escaped_char('`'),
       parse_escaped_char('"'),
@@ -516,12 +532,12 @@ pub(crate) fn parse_string_parts(
     enum PendingPart<'a> {
       Char(char),
       Variable(&'a str),
-      SubShell(SequentialList),
+      Command(SequentialList),
     }
 
     let (input, parts) = many0(or5(
       map(first_escaped_char(&allow_char), PendingPart::Char),
-      map(parse_subshell, PendingPart::SubShell),
+      map(parse_command_substitution, PendingPart::Command),
       map(
         preceded(char('$'), parse_env_var_name),
         PendingPart::Variable,
@@ -546,7 +562,7 @@ pub(crate) fn parse_string_parts(
             result.push(StringPart::Text(c.to_string()));
           }
         }
-        PendingPart::SubShell(s) => result.push(StringPart::SubShell(s)),
+        PendingPart::Command(s) => result.push(StringPart::Command(s)),
         PendingPart::Variable(v) => {
           result.push(StringPart::Variable(v.to_string()))
         }
@@ -557,10 +573,21 @@ pub(crate) fn parse_string_parts(
   }
 }
 
-fn parse_subshell(input: &str) -> ParseResult<SequentialList> {
+fn parse_command_substitution(input: &str) -> ParseResult<SequentialList> {
   delimited(tag("$("), parse_sequential_list, char(')'))(input)
 }
 
+fn parse_subshell(input: &str) -> ParseResult<SequentialList> {
+  delimited(
+    terminated(char('('), skip_whitespace),
+    parse_sequential_list,
+    with_failure_input(
+      input,
+      assert_exists(char(')'), "Expected closing parenthesis on subshell."),
+    ),
+  )(input)
+}
+
 fn parse_usize(input: &str) -> ParseResult<usize> {
   let mut value = 0;
   let mut byte_index = 0;
@@ -592,19 +619,6 @@ fn assert_whitespace_or_end(input: &str) -> ParseResult<()> {
   Ok((input, ()))
 }
 
-fn special_char(input: &str) -> ParseResult<char> {
-  if let Some((index, next_char)) = input.char_indices().next() {
-    if is_special_char(next_char) {
-      return Ok((&input[index..], next_char));
-    }
-  }
-  ParseError::backtrace()
-}
-
-fn is_special_char(c: char) -> bool {
-  "*~(){}<>$|&;\"'".contains(c)
-}
-
 fn is_valid_env_var_char(c: char) -> bool {
   // [a-zA-Z0-9_]+
   c.is_ascii_alphanumeric() || c == '_'
@@ -631,11 +645,17 @@ fn is_reserved_word(text: &str) -> bool {
 
 fn fail_for_trailing_input(input: &str) -> ParseErrorFailure {
   if parse_redirect(input).is_ok() {
-    ParseErrorFailure::new(input, "Redirects are currently not supported.")
+    ParseErrorFailure::new(
+      input,
+      "Redirects are currently not supported, but will be soon.",
+    )
   } else if input.starts_with('*') {
-    ParseErrorFailure::new(input, "Globs are currently not supported.")
+    ParseErrorFailure::new(
+      input,
+      "Globs are currently not supported, but will be soon.",
+    )
   } else {
-    ParseErrorFailure::new(input, "Unsupported character.")
+    ParseErrorFailure::new(input, "Unexpected character.")
   }
 }
 
@@ -649,31 +669,58 @@ mod test {
     assert_eq!(parse("").err().unwrap().to_string(), "Empty command.");
     assert_eq!(
       parse("&& testing").err().unwrap().to_string(),
-      concat!("Unsupported character.\n", "  && testing\n", "  ~",),
+      concat!("Unexpected character.\n", "  && testing\n", "  ~",),
     );
     assert_eq!(
       parse("test { test").err().unwrap().to_string(),
-      concat!("Unsupported character.\n", "  { test\n", "  ~",),
+      concat!("Unexpected character.\n", "  { test\n", "  ~",),
     );
     assert_eq!(
       parse("test > redirect").err().unwrap().to_string(),
       concat!(
-        "Redirects are currently not supported.\n",
+        "Redirects are currently not supported, but will be soon.\n",
         "  > redirect\n",
         "  ~",
       ),
     );
     assert_eq!(
       parse("cp test/* other").err().unwrap().to_string(),
-      concat!("Globs are currently not supported.\n", "  * other\n", "  ~",),
+      concat!(
+        "Globs are currently not supported, but will be soon.\n",
+        "  * other\n",
+        "  ~",
+      ),
     );
+    assert_eq!(
+      parse("(test").err().unwrap().to_string(),
+      concat!(
+        "Expected closing parenthesis on subshell.\n",
+        "  (test\n",
+        "  ~"
+      ),
+    );
+    assert_eq!(
+      parse("cmd \"test").err().unwrap().to_string(),
+      concat!("Expected closing double quote.\n", "  \"test\n", "  ~"),
+    );
+    assert_eq!(
+      parse("cmd 'test").err().unwrap().to_string(),
+      concat!("Expected closing single quote.\n", "  'test\n", "  ~"),
+    );
+
+    assert!(parse("( test ||other&&test;test);(t&est );").is_ok());
   }
 
   #[test]
   fn test_sequential_list() {
     run_test(
       parse_sequential_list,
-      "Name=Value OtherVar=Other command arg1 || command2 arg12 arg13 ; command3 && command4 & command5 ; export ENV6=5 ; ENV7=other && command8 || command9",
+      concat!(
+        "Name=Value OtherVar=Other command arg1 || command2 arg12 arg13 ; ",
+        "command3 && command4 & command5 ; export ENV6=5 ; ",
+        "ENV7=other && command8 || command9 ; ",
+        "cmd10 && (cmd11 || cmd12)"
+      ),
       Ok(SequentialList {
         items: vec![
           SequentialListItem {
@@ -681,10 +728,19 @@ mod test {
             sequence: Sequence::BooleanList(Box::new(BooleanList {
               current: Sequence::Command(Command {
                 env_vars: vec![
-                  EnvVar::new("Name".to_string(), StringOrWord::new_word("Value")),
-                  EnvVar::new("OtherVar".to_string(), StringOrWord::new_word("Other")),
+                  EnvVar::new(
+                    "Name".to_string(),
+                    StringOrWord::new_word("Value"),
+                  ),
+                  EnvVar::new(
+                    "OtherVar".to_string(),
+                    StringOrWord::new_word("Other"),
+                  ),
+                ],
+                args: vec![
+                  StringOrWord::new_word("command"),
+                  StringOrWord::new_word("arg1"),
                 ],
-                args: vec![StringOrWord::new_word("command"), StringOrWord::new_word("arg1")],
               }),
               op: BooleanListOperator::Or,
               next: Sequence::Command(Command {
@@ -720,12 +776,18 @@ mod test {
           },
           SequentialListItem {
             is_async: false,
-            sequence: Sequence::EnvVar(EnvVar::new("ENV6".to_string(), StringOrWord::new_word("5"))),
+            sequence: Sequence::EnvVar(EnvVar::new(
+              "ENV6".to_string(),
+              StringOrWord::new_word("5"),
+            )),
           },
           SequentialListItem {
             is_async: false,
             sequence: Sequence::BooleanList(Box::new(BooleanList {
-              current: Sequence::ShellVar(EnvVar::new("ENV7".to_string(), StringOrWord::new_word("other"))),
+              current: Sequence::ShellVar(EnvVar::new(
+                "ENV7".to_string(),
+                StringOrWord::new_word("other"),
+              )),
               op: BooleanListOperator::And,
               next: Sequence::BooleanList(Box::new(BooleanList {
                 current: Sequence::Command(Command {
@@ -740,8 +802,34 @@ mod test {
               })),
             })),
           },
+          SequentialListItem {
+            is_async: false,
+            sequence: Sequence::BooleanList(Box::new(BooleanList {
+              current: Sequence::Command(Command {
+                env_vars: vec![],
+                args: vec![StringOrWord::new_word("cmd10")],
+              }),
+              op: BooleanListOperator::And,
+              next: Sequence::Subshell(Box::new(SequentialList {
+                items: vec![SequentialListItem {
+                  is_async: false,
+                  sequence: Sequence::BooleanList(Box::new(BooleanList {
+                    current: Sequence::Command(Command {
+                      env_vars: vec![],
+                      args: vec![StringOrWord::new_word("cmd11")],
+                    }),
+                    op: BooleanListOperator::Or,
+                    next: Sequence::Command(Command {
+                      env_vars: vec![],
+                      args: vec![StringOrWord::new_word("cmd12")],
+                    }),
+                  })),
+                }],
+              })),
+            })),
+          },
         ],
-      })
+      }),
     );
 
     run_test(
@@ -884,7 +972,7 @@ mod test {
       "Name=$(test)",
       Ok(EnvVar {
         name: "Name".to_string(),
-        value: StringOrWord::Word(vec![StringPart::SubShell(SequentialList {
+        value: StringOrWord::Word(vec![StringPart::Command(SequentialList {
           items: vec![SequentialListItem {
             is_async: false,
             sequence: Sequence::Command(Command {
@@ -901,7 +989,7 @@ mod test {
       "Name=$(OTHER=5)",
       Ok(EnvVar {
         name: "Name".to_string(),
-        value: StringOrWord::Word(vec![StringPart::SubShell(SequentialList {
+        value: StringOrWord::Word(vec![StringPart::Command(SequentialList {
           items: vec![SequentialListItem {
             is_async: false,
             sequence: Sequence::ShellVar(EnvVar {
@@ -999,6 +1087,12 @@ mod test {
   #[test]
   fn test_parse_word() {
     run_test(parse_word, "if", Err("Unsupported reserved word."));
+    run_test(parse_word, "$", Ok(vec![StringPart::Text("$".to_string())]));
+    // unsupported shell variables
+    run_test(parse_word, "$$", Err("$$ is currently not supported."));
+    run_test(parse_word, "$?", Err("$? is currently not supported."));
+    run_test(parse_word, "$#", Err("$# is currently not supported."));
+    run_test(parse_word, "$*", Err("$* is currently not supported."));
   }
 
   #[test]
