@@ -1,6 +1,7 @@
 // Copyright 2018-2022 the Deno authors. All rights reserved. MIT license.
 
 use std::collections::HashMap;
+use std::path::Path;
 use std::path::PathBuf;
 use std::process::Stdio;
 
@@ -13,70 +14,90 @@ use tokio::io::AsyncWriteExt;
 
 use crate::commands::cd_command;
 use crate::commands::cp_command;
+use crate::commands::exit_command;
+use crate::commands::mkdir_command;
 use crate::commands::mv_command;
 use crate::commands::rm_command;
 use crate::commands::sleep_command;
+use crate::environment::Environment;
+use crate::environment::RealEnvironment;
 use crate::parser::Command;
 use crate::parser::Sequence;
 use crate::parser::SequentialList;
 use crate::parser::StringOrWord;
 use crate::parser::StringPart;
 use crate::shell_types::EnvChange;
-use crate::shell_types::EnvState;
 use crate::shell_types::ExecuteResult;
-use crate::shell_types::ExecutedTask;
 use crate::shell_types::ShellPipe;
 use crate::shell_types::ShellPipeSender;
+use crate::shell_types::ShellState;
+use crate::shell_types::SpawnedStep;
 
 pub async fn execute(
   list: SequentialList,
   env_vars: HashMap<String, String>,
-  cwd: PathBuf,
-  additional_cli_args: Vec<String>,
-) -> Result<i32> {
-  assert!(cwd.is_absolute());
-  let list = append_cli_args(list, additional_cli_args)?;
-  let state = EnvState::new(env_vars, cwd);
-  let executed = execute_sequential_list(list, state, ShellPipe::InheritStdin);
-  // todo: something better at outputting to stdout?
-  let output_task = tokio::task::spawn(async move {
-    executed.stdout.pipe_to_stdout().await;
-  });
-  let result = executed.task.await;
-  output_task.await.unwrap();
-
-  Ok(match result {
-    ExecuteResult::Exit => 1,
-    ExecuteResult::Continue(exit_code, _) => exit_code,
-  })
+  cwd: &Path,
+) -> i32 {
+  let environment = RealEnvironment::default();
+  execute_with_environment(list, env_vars, cwd, environment).await
 }
 
-/// When a user calls `deno task <task-name> -- <args>`, we want
-/// to append those CLI arguments to the last command.
-fn append_cli_args(
-  mut list: SequentialList,
-  args: Vec<String>,
-) -> Result<SequentialList> {
-  if args.is_empty() {
-    return Ok(list);
-  }
-
-  // todo(THIS PR): this part and remove this clippy
-  #[allow(clippy::redundant_pattern_matching)]
-  if let Some(_) = list.items.last_mut() {
-    todo!();
-  }
-
-  Ok(list)
-}
-
-fn execute_sequential_list(
+pub(crate) async fn execute_with_environment(
   list: SequentialList,
-  mut state: EnvState,
+  env_vars: HashMap<String, String>,
+  cwd: &Path,
+  environment: impl Environment,
+) -> i32 {
+  assert!(cwd.is_absolute());
+  let (env_stdout_tx, env_stdout) = ShellPipe::channel();
+  let state = ShellState::new(env_vars, cwd, env_stdout_tx.clone());
+
+  // spawn a task to pipe env_stdout to the actual environment's stdout
+  let env_stdout_output_handle = tokio::task::spawn({
+    let environment = environment.clone();
+    async move {
+      env_stdout.pipe_to_writer(environment.async_stdout()).await;
+    }
+  });
+
+  // spawn a sequential list and pipe its output to the environment
+  let spawned_step = spawn_sequential_list(
+    list,
+    state,
+    ShellPipe::InheritStdin,
+    environment.clone(),
+    AsyncCommandBehavior::Wait,
+  );
+  let result = spawned_step
+    .wait_with_stdout(move |stdout| async move {
+      stdout.pipe_to_sender(env_stdout_tx).await;
+    })
+    .await;
+
+  // wait on the environment's stdout handle to finish draining
+  env_stdout_output_handle.await.unwrap();
+
+  match result {
+    ExecuteResult::Exit(code, _) => code,
+    ExecuteResult::Continue(exit_code, _, _) => exit_code,
+  }
+}
+
+#[derive(Debug, PartialEq)]
+enum AsyncCommandBehavior {
+  Wait,
+  Yield,
+}
+
+fn spawn_sequential_list(
+  list: SequentialList,
+  mut state: ShellState,
   mut stdin: ShellPipe,
-) -> ExecutedTask {
+  environment: impl Environment,
+  async_command_behavior: AsyncCommandBehavior,
+) -> SpawnedStep {
   let (stdout_tx, stdout) = ShellPipe::channel();
-  ExecutedTask {
+  SpawnedStep {
     stdout,
     task: async move {
       let mut final_exit_code = 0;
@@ -85,44 +106,48 @@ fn execute_sequential_list(
       for item in list.items {
         if item.is_async {
           let state = state.clone();
-          let stdout_tx = stdout_tx.clone();
+          let environment = environment.clone();
           async_handles.push(tokio::task::spawn(async move {
-            let command = execute_sequence(
+            // use the current shell's stdout_tx in order to avoid
+            // blocking async out
+            let shell_stdout_tx = state.shell_stdout_tx();
+            let spawned_step = spawn_sequence(
               item.sequence,
               state,
-              // todo: not correct... should use provided stdin
+              // todo(#2): not correct... should use provided stdin
               ShellPipe::InheritStdin,
+              environment,
             )
             .await;
-            // todo: better code
-            let output_task = tokio::task::spawn(async move {
-              command.stdout.pipe_to_sender(stdout_tx).await;
-            });
-            let result = command.task.await;
-            output_task.await.unwrap();
-            result
+            let result = spawned_step
+              .wait_with_stdout(move |stdout| async move {
+                stdout.pipe_to_sender(shell_stdout_tx).await;
+              })
+              .await;
+            futures::future::join_all(result.into_handles()).await;
           }));
         } else {
-          // todo: this doesn't seem correct. I believe all these commands
-          // should use the same stdin and not inherit from the process after
-          // the first item in the sequential list.
+          // todo(#2): not correct... should share the provided stdin
           let stdin = std::mem::replace(&mut stdin, ShellPipe::InheritStdin);
-          let command =
-            execute_sequence(item.sequence, state.clone(), stdin).await;
-          // todo: something better?
-          let output_task = tokio::task::spawn({
-            let stdout_tx = stdout_tx.clone();
-            async move {
-              command.stdout.pipe_to_sender(stdout_tx).await;
-            }
-          });
-          let result = command.task.await;
-          output_task.await.unwrap();
+          let command = spawn_sequence(
+            item.sequence,
+            state.clone(),
+            stdin,
+            environment.clone(),
+          )
+          .await;
+          let stdout_tx = stdout_tx.clone();
+          let result = command
+            .wait_with_stdout(move |stdout| async {
+              stdout.pipe_to_sender(stdout_tx).await;
+            })
+            .await;
           match result {
-            ExecuteResult::Exit => return ExecuteResult::Exit,
-            ExecuteResult::Continue(exit_code, changes) => {
+            ExecuteResult::Exit(_, _) => return result,
+            ExecuteResult::Continue(exit_code, changes, handles) => {
               state.apply_changes(&changes);
               final_changes.extend(changes);
+              async_handles.extend(handles);
               // use the final sequential item's exit code
               final_exit_code = exit_code;
             }
@@ -131,62 +156,72 @@ fn execute_sequential_list(
       }
 
       // wait for async commands to complete
-      futures::future::join_all(async_handles).await;
+      if async_command_behavior == AsyncCommandBehavior::Wait {
+        futures::future::join_all(async_handles.drain(..)).await;
+      }
 
-      ExecuteResult::Continue(final_exit_code, final_changes)
+      ExecuteResult::Continue(final_exit_code, final_changes, async_handles)
     }
     .boxed(),
   }
 }
 
-// todo(THIS PR): clean up this function
-fn execute_sequence(
+fn spawn_sequence(
   sequence: Sequence,
-  mut state: EnvState,
+  mut state: ShellState,
   stdin: ShellPipe,
-) -> BoxFuture<'static, ExecutedTask> {
+  environment: impl Environment,
+) -> BoxFuture<'static, SpawnedStep> {
   // requires boxed async because of recursive async
   async move {
     match sequence {
       Sequence::EnvVar(var) => {
-        ExecutedTask::from_result(ExecuteResult::Continue(
+        SpawnedStep::from_result(ExecuteResult::Continue(
           0,
           vec![EnvChange::SetEnvVar(
             var.name,
-            evaluate_string_or_word(var.value, &state).await,
+            evaluate_string_or_word(var.value, &state, environment).await,
           )],
+          Vec::new(),
         ))
       }
       Sequence::ShellVar(var) => {
-        ExecutedTask::from_result(ExecuteResult::Continue(
+        SpawnedStep::from_result(ExecuteResult::Continue(
           0,
           vec![EnvChange::SetShellVar(
             var.name,
-            evaluate_string_or_word(var.value, &state).await,
+            evaluate_string_or_word(var.value, &state, environment).await,
           )],
+          Vec::new(),
         ))
       }
-      Sequence::Command(command) => start_command(command, &state, stdin).await,
+      Sequence::Command(command) => {
+        start_command(command, &state, stdin, environment).await
+      }
       Sequence::BooleanList(list) => {
         let (stdout_tx, stdout) = ShellPipe::channel();
-        ExecutedTask {
+        SpawnedStep {
           stdout,
           task: async move {
-            // todo(THIS PR): clean this up
             let mut changes = vec![];
-            let first_result = execute_and_wait_sequence(
+            let first_result = spawn_and_wait_sequence(
               list.current,
               state.clone(),
               stdin,
               stdout_tx.clone(),
+              environment.clone(),
             )
             .await;
-            let exit_code = match first_result {
-              ExecuteResult::Exit => return ExecuteResult::Exit,
-              ExecuteResult::Continue(exit_code, sub_changes) => {
+            let (exit_code, mut async_handles) = match first_result {
+              ExecuteResult::Exit(_, _) => return first_result,
+              ExecuteResult::Continue(
+                exit_code,
+                sub_changes,
+                async_handles,
+              ) => {
                 state.apply_changes(&sub_changes);
                 changes.extend(sub_changes);
-                exit_code
+                (exit_code, async_handles)
               }
             };
 
@@ -208,23 +243,32 @@ fn execute_sequence(
               }
             };
             if let Some(next) = next {
-              let next_result = execute_and_wait_sequence(
+              let next_result = spawn_and_wait_sequence(
                 next,
                 state.clone(),
-                // seems suspect, but good enough for now
+                // todo(#2): share stdin
                 ShellPipe::InheritStdin,
                 stdout_tx.clone(),
+                environment,
               )
               .await;
               match next_result {
-                ExecuteResult::Exit => ExecuteResult::Exit,
-                ExecuteResult::Continue(exit_code, sub_changes) => {
+                ExecuteResult::Exit(code, sub_handles) => {
+                  async_handles.extend(sub_handles);
+                  ExecuteResult::Exit(code, async_handles)
+                }
+                ExecuteResult::Continue(
+                  exit_code,
+                  sub_changes,
+                  sub_handles,
+                ) => {
                   changes.extend(sub_changes);
-                  ExecuteResult::Continue(exit_code, changes)
+                  async_handles.extend(sub_handles);
+                  ExecuteResult::Continue(exit_code, changes, async_handles)
                 }
               }
             } else {
-              ExecuteResult::Continue(exit_code, changes)
+              ExecuteResult::Continue(exit_code, changes, async_handles)
             }
           }
           .boxed(),
@@ -232,36 +276,72 @@ fn execute_sequence(
       }
       Sequence::Pipeline(pipeline) => {
         let (stdout_tx, stdout) = ShellPipe::channel();
-        ExecutedTask {
+        SpawnedStep {
           stdout,
           task: async move {
             let sequences = pipeline.into_vec();
             let mut wait_tasks = vec![];
             let mut last_input = Some(stdin);
             for sequence in sequences.into_iter() {
-              let executed_sequence = execute_sequence(
+              let spawned_sequence = spawn_sequence(
                 sequence,
                 state.clone(),
                 last_input.take().unwrap(),
+                environment.clone(),
               )
               .await;
-              last_input = Some(executed_sequence.stdout);
-              wait_tasks.push(executed_sequence.task);
+              last_input = Some(spawned_sequence.stdout);
+              wait_tasks.push(spawned_sequence.task);
             }
-            // todo: something better
-            let output_task = tokio::task::spawn({
+            let output_handle = tokio::task::spawn({
               async move {
                 last_input.unwrap().pipe_to_sender(stdout_tx).await;
               }
             });
             let mut results = futures::future::join_all(wait_tasks).await;
-            output_task.await.unwrap();
+            output_handle.await.unwrap();
             let last_result = results.pop().unwrap();
+            let all_handles =
+              results.into_iter().flat_map(|r| r.into_handles());
             match last_result {
-              ExecuteResult::Exit => ExecuteResult::Continue(1, Vec::new()),
-              ExecuteResult::Continue(exit_code, _) => {
-                ExecuteResult::Continue(exit_code, Vec::new())
+              ExecuteResult::Exit(code, mut handles) => {
+                handles.extend(all_handles);
+                ExecuteResult::Continue(code, Vec::new(), handles)
               }
+              ExecuteResult::Continue(code, _, mut handles) => {
+                handles.extend(all_handles);
+                ExecuteResult::Continue(code, Vec::new(), handles)
+              }
+            }
+          }
+          .boxed(),
+        }
+      }
+      Sequence::Subshell(list) => {
+        let (stdout_tx, stdout) = ShellPipe::channel();
+        SpawnedStep {
+          stdout,
+          task: async move {
+            let step = spawn_sequential_list(
+              *list,
+              state.clone(),
+              stdin,
+              environment,
+              // yield async commands to the parent
+              AsyncCommandBehavior::Yield,
+            );
+            let result = step
+              .wait_with_stdout(move |stdout| async move {
+                stdout.pipe_to_sender(stdout_tx).await;
+              })
+              .await;
+
+            // sub shells do not cause an exit
+            match result {
+              ExecuteResult::Exit(code, handles) => {
+                ExecuteResult::Continue(code, Vec::new(), handles)
+              }
+              ExecuteResult::Continue(_, _, _) => result,
             }
           }
           .boxed(),
@@ -272,31 +352,28 @@ fn execute_sequence(
   .boxed()
 }
 
-async fn execute_and_wait_sequence(
+async fn spawn_and_wait_sequence(
   sequence: Sequence,
-  state: EnvState,
+  state: ShellState,
   stdin: ShellPipe,
   sender: ShellPipeSender,
+  environment: impl Environment,
 ) -> ExecuteResult {
-  let command = execute_sequence(sequence, state, stdin).await;
-  // todo: something better
-  let output_task = tokio::task::spawn({
-    async move {
-      command.stdout.pipe_to_sender(sender).await;
-    }
-  });
-  let result = command.task.await;
-  output_task.await.unwrap();
-  result
+  let step = spawn_sequence(sequence, state, stdin, environment).await;
+  step
+    .wait_with_stdout(move |stdout| async {
+      stdout.pipe_to_sender(sender).await;
+    })
+    .await
 }
 
 async fn start_command(
   command: Command,
-  state: &EnvState,
+  state: &ShellState,
   stdin: ShellPipe,
-) -> ExecutedTask {
-  // todo: reduce code duplication in here
-  let mut args = evaluate_args(command.args, state).await;
+  environment: impl Environment,
+) -> SpawnedStep {
+  let mut args = evaluate_args(command.args, state, environment.clone()).await;
   let command_name = if args.is_empty() {
     String::new()
   } else {
@@ -304,105 +381,62 @@ async fn start_command(
   };
   if command_name == "cd" {
     let cwd = state.cwd().clone();
-    let (tx, stdout) = ShellPipe::channel();
-    ExecutedTask {
-      stdout,
-      task: async move {
-        let result = cd_command(&cwd, args);
-        drop(tx); // close stdout
-        result
-      }
-      .boxed(),
-    }
+    SpawnedStep::from_fn(move || cd_command(&cwd, args, environment))
   } else if command_name == "exit" {
-    let (tx, stdout) = ShellPipe::channel();
-    ExecutedTask {
-      stdout,
-      task: async move {
-        drop(tx); // close stdout
-        if !args.is_empty() {
-          eprintln!("exit had too many arguments.");
-          ExecuteResult::Continue(1, Vec::new())
-        } else {
-          ExecuteResult::Exit
-        }
-      }
-      .boxed(),
-    }
+    SpawnedStep::from_future(async move { exit_command(args, environment) })
   } else if command_name == "pwd" {
     // ignores additional arguments
-    ExecutedTask::with_stdout_text(format!("{}\n", state.cwd().display()))
+    SpawnedStep::with_stdout_text(format!("{}\n", state.cwd().display()))
   } else if command_name == "echo" {
-    ExecutedTask::with_stdout_text(format!("{}\n", args.join(" ")))
+    SpawnedStep::with_stdout_text(format!("{}\n", args.join(" ")))
   } else if command_name == "true" {
     // ignores additional arguments
-    ExecutedTask::from_exit_code(0)
+    SpawnedStep::from_exit_code(0)
   } else if command_name == "false" {
     // ignores additional arguments
-    ExecutedTask::from_exit_code(1)
+    SpawnedStep::from_exit_code(1)
   } else if command_name == "cp" {
     let cwd = state.cwd().clone();
-    let (tx, stdout) = ShellPipe::channel();
-    ExecutedTask {
-      stdout,
-      task: async move {
-        let result = cp_command(&cwd, args).await;
-        drop(tx); // close stdout
-        result
-      }
-      .boxed(),
-    }
+    SpawnedStep::from_future(async move {
+      cp_command(&cwd, args, environment).await
+    })
+  } else if command_name == "mkdir" {
+    let cwd = state.cwd().clone();
+    SpawnedStep::from_future(async move {
+      mkdir_command(&cwd, args, environment).await
+    })
   } else if command_name == "mv" {
     let cwd = state.cwd().clone();
-    let (tx, stdout) = ShellPipe::channel();
-    ExecutedTask {
-      stdout,
-      task: async move {
-        let result = mv_command(&cwd, args).await;
-        drop(tx); // close stdout
-        result
-      }
-      .boxed(),
-    }
+    SpawnedStep::from_future(async move {
+      mv_command(&cwd, args, environment).await
+    })
   } else if command_name == "rm" {
     let cwd = state.cwd().clone();
-    let (tx, stdout) = ShellPipe::channel();
-    ExecutedTask {
-      stdout,
-      task: async move {
-        let result = rm_command(&cwd, args).await;
-        drop(tx); // close stdout
-        result
-      }
-      .boxed(),
-    }
+    SpawnedStep::from_future(async move {
+      rm_command(&cwd, args, environment).await
+    })
   } else if command_name == "sleep" {
-    let (tx, stdout) = ShellPipe::channel();
-    ExecutedTask {
-      stdout,
-      task: async move {
-        let result = sleep_command(args).await;
-        drop(tx); // close stdout
-        result
-      }
-      .boxed(),
-    }
+    SpawnedStep::from_future(
+      async move { sleep_command(args, environment).await },
+    )
   } else {
     let mut state = state.clone();
     for env_var in command.env_vars {
       state.apply_env_var(
         &env_var.name,
-        &evaluate_string_or_word(env_var.value, &state).await,
+        &evaluate_string_or_word(env_var.value, &state, environment.clone())
+          .await,
       );
     }
 
     let command_path = match resolve_command_path(&command_name, &state).await {
       Ok(command_path) => command_path,
       Err(err) => {
-        eprintln!("Error launching '{}': {}", command_name, err);
-        return ExecutedTask::from_result(ExecuteResult::Continue(
+        environment.eprintln(&err.to_string());
+        return SpawnedStep::from_result(ExecuteResult::Continue(
           1,
           Vec::new(),
+          Vec::new(),
         ));
       }
     };
@@ -423,10 +457,12 @@ async fn start_command(
     let mut child = match child {
       Ok(child) => child,
       Err(err) => {
-        eprintln!("Error launching '{}': {}", command_name, err);
-        return ExecutedTask::from_result(ExecuteResult::Continue(
+        environment
+          .eprintln(&format!("Error launching '{}': {}", command_name, err));
+        return SpawnedStep::from_result(ExecuteResult::Continue(
           1,
           Vec::new(),
+          Vec::new(),
         ));
       }
     };
@@ -447,7 +483,7 @@ async fn start_command(
     let mut child_stdout = child.stdout.take().unwrap();
     let (stdout_tx, stdout) = ShellPipe::channel();
 
-    ExecutedTask {
+    SpawnedStep {
       stdout,
       task: async move {
         // spawn a task to pipe the messages from the process' stdout to the channel
@@ -470,13 +506,14 @@ async fn start_command(
         });
 
         match child.wait().await {
-          Ok(status) => {
-            // TODO(THIS PR): Is unwrapping to 1 ok here?
-            ExecuteResult::Continue(status.code().unwrap_or(1), Vec::new())
-          }
+          Ok(status) => ExecuteResult::Continue(
+            status.code().unwrap_or(1),
+            Vec::new(),
+            Vec::new(),
+          ),
           Err(err) => {
-            eprintln!("{}", err);
-            ExecuteResult::Continue(1, Vec::new())
+            environment.eprintln(&format!("{}", err));
+            ExecuteResult::Continue(1, Vec::new(), Vec::new())
           }
         }
       }
@@ -487,10 +524,10 @@ async fn start_command(
 
 async fn resolve_command_path(
   command_name: &str,
-  state: &EnvState,
+  state: &ShellState,
 ) -> Result<PathBuf> {
   if command_name.is_empty() {
-    bail!("Command name was empty.");
+    bail!("command name was empty");
   }
 
   // check for absolute
@@ -549,20 +586,22 @@ async fn resolve_command_path(
     }
   }
 
-  bail!("Command not found.")
+  bail!("{}: command not found", command_name)
 }
 
 async fn evaluate_args(
   args: Vec<StringOrWord>,
-  state: &EnvState,
+  state: &ShellState,
+  environment: impl Environment,
 ) -> Vec<String> {
   let mut result = Vec::new();
   for arg in args {
     match arg {
       StringOrWord::Word(parts) => {
-        // todo: maybe we should have this work like sh and I believe
-        // reparse then continually re-evaluate until there's only strings left
-        let text = evaluate_string_parts(parts, state).await;
+        // todo(dsherret): maybe we should have this work like sh and I believe
+        // reparse then continually re-evaluate until there's only strings left.
+        let text =
+          evaluate_string_parts(parts, state, environment.clone()).await;
         for part in text.split(' ') {
           let part = part.trim();
           if !part.is_empty() {
@@ -571,7 +610,8 @@ async fn evaluate_args(
         }
       }
       StringOrWord::String(parts) => {
-        result.push(evaluate_string_parts(parts, state).await);
+        result
+          .push(evaluate_string_parts(parts, state, environment.clone()).await);
       }
     }
   }
@@ -580,14 +620,16 @@ async fn evaluate_args(
 
 async fn evaluate_string_or_word(
   string_or_word: StringOrWord,
-  state: &EnvState,
+  state: &ShellState,
+  environment: impl Environment,
 ) -> String {
-  evaluate_string_parts(string_or_word.into_parts(), state).await
+  evaluate_string_parts(string_or_word.into_parts(), state, environment).await
 }
 
 async fn evaluate_string_parts(
   parts: Vec<StringPart>,
-  state: &EnvState,
+  state: &ShellState,
+  environment: impl Environment,
 ) -> String {
   let mut final_text = String::new();
   for part in parts {
@@ -598,39 +640,58 @@ async fn evaluate_string_parts(
           final_text.push_str(value);
         }
       }
-      StringPart::SubShell(list) => {
-        final_text.push_str(&evaluate_string_part_subshell(list, state).await)
-      }
+      StringPart::Command(list) => final_text.push_str(
+        &evaluate_command_substitution(list, state, environment.clone()).await,
+      ),
     }
   }
   final_text
 }
 
-async fn evaluate_string_part_subshell(
+async fn evaluate_command_substitution(
   list: SequentialList,
-  state: &EnvState,
+  state: &ShellState,
+  environment: impl Environment,
 ) -> String {
-  let task = execute_sequential_list(
+  let (shell_tx, shell_stdout) = ShellPipe::channel();
+  let spawned_step = spawn_sequential_list(
     list,
-    state.clone(),
-    // todo: this is not correct. It should use the current stdin.
+    {
+      // command substitution needs to capture even the async
+      // output, so set the current state's shell_stdout_tx
+      // in order to capture it
+      let mut state = state.clone();
+      state.set_shell_stdout_tx(shell_tx);
+      state
+    },
+    // todo(#2): this is not correct. It should use the current stdin.
     ShellPipe::InheritStdin,
+    environment,
+    AsyncCommandBehavior::Wait,
   );
-  // todo: something better?
-  let output_task = tokio::task::spawn(async {
+  let output_handle = tokio::task::spawn(async {
     let mut final_data = Vec::new();
-    match task.stdout {
-      ShellPipe::InheritStdin => unreachable!(),
-      ShellPipe::Channel(mut rx) => {
-        while let Some(data) = rx.recv().await {
+    let mut shell_rx = shell_stdout.unwrap_channel();
+    let mut spawned_step_rx = spawned_step.stdout.unwrap_channel();
+
+    loop {
+      tokio::select! {
+        Some(data) = shell_rx.recv() => {
           final_data.extend(data);
         }
+        Some(data) = spawned_step_rx.recv() => {
+          final_data.extend(data);
+        }
+        else => {
+          break;
+        }
       }
     }
+
     final_data
   });
-  let _ = task.task.await;
-  let data = output_task.await.unwrap();
+  let _ = spawned_step.task.await;
+  let data = output_handle.await.unwrap();
   let text = String::from_utf8_lossy(&data);
 
   // Remove the trailing newline and then replace inner newlines with a space
